"""
skill_tagging Django application initialization.
"""

from django.apps import AppConfig


class SkillTaggingConfig(AppConfig):
    """
    Configuration for the skill_tagging Django application.
    """

    name = 'skill_tagging'

    plugin_app = {
        "signals_config": {
            "lms.djangoapp": {
                "relative_path": "handlers",
                "receivers": [
                    {
                        "receiver_func_name": "listen_for_xblock_skill_verified",
                        "signal_path": "openedx_events.learning.signals.XBLOCK_SKILL_VERIFIED",
                    },
                ],
            }
        },
    }



"""
Module that contains the openedx_filters pipeline steps.
"""
import logging
import random

import pkg_resources
from django.conf import settings
from django.template import Context, Template
from openedx_filters import PipelineStep

logger = logging.getLogger(__name__)
DEFAULT_PROBABILITY = 0.5


def resource_string(path):
    """Handy helper for getting resources from our kit."""
    data = pkg_resources.resource_string(__name__, path)
    return data.decode("utf8")


class VerificationPipelineBase:
    """
    Common functions for verification pipelines.
    """
    @staticmethod
    def fetch_related_skills(block):
        """Checks `has_verified_tags` and fetchs related skills."""
        has_verified_tags = getattr(block, "has_verified_tags", None)
        if has_verified_tags is None or has_verified_tags is True:
            return []
        fetch_tags = getattr(block, "fetch_skill_tags", None)
        if fetch_tags is None:
            return []
        tags = fetch_tags()
        return tags

    @staticmethod
    def should_run_filter():
        """Determines whether we should run filter and display form."""
        # random returns a number between 0 and 1 (inclusive).
        probability = getattr(settings, "SHOW_SKILL_VERIFICATION_PROBABILITY", DEFAULT_PROBABILITY)
        return random.random() < probability

    @staticmethod
    def get_skill_context(usage_id, block, skills):
        """
        Build common skill context data used by templates.
        """
        verify_tags_url = block.runtime.handler_url(block, "verify_tags")
        data = {
            "block_type": usage_id.block_type,
            "block_id": usage_id.block_id,
            "verify_tags_url": verify_tags_url,
            "skills": skills,
            "tag_verification_template": Template(resource_string("static/tagging.html")),
            "form": Template(resource_string("static/tags_form.html")),
            "thank_you_section": Template(resource_string("static/tags_thankyou.html")),
            "error_section": Template(resource_string("static/tags_error.html")),
            "image": resource_string("static/brainstorming.svg"),
        }
        return data


class AddVerticalBlockSkillVerificationSection(VerificationPipelineBase, PipelineStep):
    """
    Adds extra HTML to the fragment.

    Example Usage:

    .. code-block::

        "OPENEDX_FILTERS_CONFIG": {
            "org.openedx.learning.vertical_block.render.completed.v1": {
                "fail_sliently": false,
                "pipeline": [
                    "skill_tagging.pipeline.AddVerticalBlockSkillVerificationSection"
                ]
            }
        }
    """
    def run_filter(self, block, fragment, context, view):  # pylint: disable=arguments-differ
        """Pipeline Step implementing the Filter"""

        # Check whether we need to run this filter and only call the API.
        if not self.should_run_filter():
            return {"block": block, "fragment": fragment, "context": context, "view": view}
        skills = self.fetch_related_skills(block)
        if not skills:
            return {"block": block, "fragment": fragment, "context": context, "view": view}
        usage_id = block.scope_ids.usage_id
        data = self.get_skill_context(usage_id, block, skills)
        html = resource_string("static/tagging.html")
        css = resource_string("static/tagging.css")
        js = resource_string("static/tagging.js")
        fragment.add_content(Template(html).render(Context(data)))
        fragment.add_javascript(js)
        fragment.add_css(css)
        return {"block": block, "fragment": fragment, "context": context, "view": view}


class AddVideoBlockSkillVerificationComponent(VerificationPipelineBase, PipelineStep):
    """
    Adds verification component to video blocks.

    Example Usage:

    .. code-block::

        "OPENEDX_FILTERS_CONFIG": {
            "org.openedx.learning.vertical_block_child.render.started.v1": {
                "fail_sliently": false,
                "pipeline": [
                    "skill_tagging.pipeline.AddVideoBlockSkillVerificationComponent"
                ]
            }
        }
    """
    def run_filter(self, block, context):  # pylint: disable=arguments-differ
        """Pipeline Step implementing the Filter"""
        usage_id = block.scope_ids.usage_id
        if usage_id.block_type != "video" or not self.should_run_filter():
            # avoid fetching skills for other xblocks
            return {"block": block, "context": context}
        skills = self.fetch_related_skills(block)
        if not skills:
            return {"block": block, "context": context}
        data = self.get_skill_context(usage_id, block, skills)

        def wrapper(fn):
            def wrapped(_context):
                fragment = fn(_context)
                js = resource_string("static/tagging.js")
                css = resource_string("static/tagging.css")
                fragment.add_javascript(Template(resource_string("static/video_tagging.js")).render(Context(data)))
                fragment.add_javascript(js)
                fragment.add_css(css)
                return fragment
            return wrapped
        block.student_view = wrapper(block.student_view)
        return {"block": block, "context": context}



"""
A mixin that fetches and verifies skills related to an Xblock,
that can be added for all XBlocks.
"""

import logging
from urllib.parse import urljoin

from django.conf import settings
from django.utils.timezone import datetime, timezone
from django.utils.translation import gettext as _
from openedx_events.learning.data import XBlockSkillVerificationData
from openedx_events.learning.signals import XBLOCK_SKILL_VERIFIED
from xblock.core import XBlock
from xblock.fields import Boolean, Scope
from xblock.internal import NamedAttributesMetaclass
from xblock.runtime import NoSuchServiceError

from .utils import get_api_client

LOGGER = logging.getLogger(__name__)
PAGE_SIZE = getattr(settings, "TAXONOMY_API_SKILL_PAGE_SIZE", 100)


# Make '_' a no-op so we can scrape strings
def _(text):
    return text


class SkillTaggingMixin(metaclass=NamedAttributesMetaclass):
    """
    XBlock Mixin for fetching and verifying skill tags
    """
    has_verified_tags = Boolean(
        display_name=_("Has verified tags"),
        default=False,
        help=_("Has user verified tags for this XBlock?"),
        scope=Scope.user_state
    )

    def _get_user_service(self):
        """
        Tries to get user service, if not found returns None.
        """
        try:
            user_service = self.runtime.service(self, 'user')
        except NoSuchServiceError:
            user_service = None
        return user_service

    def fetch_skill_tags(self):
        """
        Fetch skill tags for the XBlock by calling taxonomy api.
        """
        if not hasattr(settings, 'TAXONOMY_API_BASE_URL'):
            LOGGER.warning(
                "Require TAXONOMY_API_BASE_URL to be present in the settings."
            )
            return []

        user_service = self._get_user_service()
        if not user_service:
            LOGGER.info(
                "No user service available for this xblock. Cannot proceed."
            )
            return []

        user = user_service.get_user_by_anonymous_id()
        api_client = get_api_client(user=user)

        usage_id_str = str(self.scope_ids.usage_id)
        XBLOCK_SKILL_TAGS_API = urljoin(
            settings.TAXONOMY_API_BASE_URL,
            '/taxonomy/api/v1/xblocks/'
        )
        response = api_client.get(
            XBLOCK_SKILL_TAGS_API,
            params={
                "usage_key": usage_id_str,
                "page_size": PAGE_SIZE,
                "verified": False,
            },
        )
        response.raise_for_status()
        results = response.json().get("results")
        if not results:
            LOGGER.info(
                f"XBlock<{usage_id_str}> does not contain any skill",
            )
            return []
        else:
            return results[0].get('skills', [])

    @XBlock.json_handler
    def verify_tags(self, data, suffix=''):  # pylint: disable=unused-argument
        """
        Handler to verify tags
        """

        verified_skills = data.get("verified_skills", [])
        ignored_skills = data.get("ignored_skills", [])
        usage_key = str(self.scope_ids.usage_id)
        if not self.has_verified_tags and (verified_skills or ignored_skills):
            XBLOCK_SKILL_VERIFIED.send_event(
                time=datetime.now(timezone.utc),
                xblock_info=XBlockSkillVerificationData(
                    usage_key=usage_key,
                    verified_skills=verified_skills,
                    ignored_skills=ignored_skills,
                )
            )
            self.has_verified_tags = True



"""
Tests for SkillTaggingMixin
"""
from __future__ import absolute_import

import unittest
from unittest.mock import Mock, patch

from test_utils import TestCaseMixin, get_tagging_mixin


@patch('skill_tagging.skill_tagging_mixin.get_api_client')
class SkillTaggingMixinTests(TestCaseMixin, unittest.TestCase):
    """
    Test suite for SkillTaggingMixin
    """

    def test_mixin_fields(self, _):
        """
        Test for mixin field and methods
        """
        self.tagging_mixin = get_tagging_mixin()
        self.assertTrue(hasattr(self.tagging_mixin, 'has_verified_tags'))
        self.assertTrue(hasattr(self.tagging_mixin, 'fetch_skill_tags'))
        self.assertTrue(hasattr(self.tagging_mixin, 'verify_tags'))

    def test_fetch_tags(self, mock_get_api_client):
        """
        Test that fetch_tags method works as expected
        """
        api_client = self.get_mock_api_response()
        mock_get_api_client.return_value = api_client
        resp = self.block.fetch_skill_tags()
        assert mock_get_api_client.call_count == 1
        assert api_client.get.call_count == 1
        expected_response = [
            {'id': 1, 'name': 'SKILL-0'},
            {'id': 6, 'name': 'SKILL-5'}
        ]
        self.assertEqual(resp, expected_response)

    @patch('skill_tagging.skill_tagging_mixin.XBLOCK_SKILL_VERIFIED')
    def test_verify_tags(self, mock_event, _):
        """
        Test that verify_tags method works as expected
        """
        mock_event.return_value = Mock(send_event=Mock())
        selected_tags = [1, 6]
        ignored_tags = []
        self.assertFalse(self.block.has_verified_tags)
        self.call_handler(self.VERIFY_TAGS_HANDLER, data={
            "verified_skills": selected_tags,
            "ignored_skills": ignored_tags,
        })
        assert mock_event.send_event.call_count == 1
        self.assertTrue(self.block.has_verified_tags)



"""
Tests for pipeline.py
"""
from unittest.mock import patch

from django.test import TestCase, override_settings
from openedx_filters.learning.filters import VerticalBlockChildRenderStarted, VerticalBlockRenderCompleted
from workbench.runtime import Fragment

from test_utils import TestCaseMixin


@patch('skill_tagging.skill_tagging_mixin.get_api_client')
@override_settings(
    OPEN_EDX_FILTERS_CONFIG={
        "org.openedx.learning.vertical_block.render.completed.v1": {
            "fail_silently": False,
            "pipeline": [
                "skill_tagging.pipeline.AddVerticalBlockSkillVerificationSection"
            ]
        }
    },
    SHOW_SKILL_VERIFICATION_PROBABILITY=0,
)
class TestAddVerticalBlockSkillVerificationSection(TestCaseMixin, TestCase):
    """
    Testcase for the AddVerticalBlockSkillVerificationSection openedx-filters pipeline.
    """

    def setUp(self) -> None:
        super().setUp()
        self.original_fragement = Fragment(content="Some")

    @override_settings(
        SHOW_SKILL_VERIFICATION_PROBABILITY=0,
    )
    def test_pipeline_does_nothing_when_probability_set_to_zero(self, mock_get_api_client):
        """Check that pipeline is not executed if probability is set to zero"""
        mock_get_api_client.return_value = self.get_mock_api_response()
        _, fragment, _, _ = VerticalBlockRenderCompleted.run_filter(
            block=self.block, context={}, fragment=self.original_fragement, view={}
        )
        mock_get_api_client.assert_not_called()
        self.assertEqual(fragment.content, self.original_fragement.content)
        self.assertNotIn("SKILL-0", fragment.content)

    @override_settings(
        SHOW_SKILL_VERIFICATION_PROBABILITY=1,
    )
    def test_pipeline_adds_verification_form_when_probability_set_to_one(self, mock_get_api_client):
        """Check that verification div is added with skills."""
        mock_get_api_client.return_value = self.get_mock_api_response()
        _, fragment, _, _ = VerticalBlockRenderCompleted.run_filter(
            block=self.block, context={}, fragment=self.original_fragement, view={}
        )

        self.assertIn(self.original_fragement.content, fragment.content)
        self.assertIn("SKILL-0", fragment.content)
        self.assertIn("SKILL-5", fragment.content)


@patch('skill_tagging.skill_tagging_mixin.get_api_client')
@override_settings(
    OPEN_EDX_FILTERS_CONFIG={
        "org.openedx.learning.vertical_block_child.render.started.v1": {
            "fail_silently": False,
            "pipeline": [
                "skill_tagging.pipeline.AddVideoBlockSkillVerificationComponent"
            ]
        }
    },
    SHOW_SKILL_VERIFICATION_PROBABILITY=1,
)
class TestAddVideoBlockSkillVerificationComponent(TestCaseMixin, TestCase):
    """
    Testcase for the AddVerticalBlockSkillVerificationSection openedx-filters pipeline.
    """

    def setUp(self) -> None:
        self.block_type = "video"
        super().setUp()
        self.original_fragement = Fragment(content="<div class='video-player'></div>")
        self.block.fragment = self.original_fragement

    @override_settings(
        SHOW_SKILL_VERIFICATION_PROBABILITY=0,
    )
    def test_pipeline_does_nothing_when_probability_set_to_zero(self, mock_get_api_client):
        """Check that pipeline is not executed if probability is set to zero"""
        mock_get_api_client.return_value = self.get_mock_api_response()
        _, fragment, _, _ = VerticalBlockRenderCompleted.run_filter(
            block=self.block, context={}, fragment=self.original_fragement, view={}
        )
        mock_get_api_client.assert_not_called()
        self.assertEqual(fragment.content, self.original_fragement.content)
        self.assertNotIn("SKILL-0", fragment.content)

    def test_pipeline_adds_required_resources(self, mock_get_api_client):
        """Check that pipeline adds required resources."""
        mock_get_api_client.return_value = self.get_mock_api_response()
        block, context = VerticalBlockChildRenderStarted.run_filter(block=self.block, context={})
        fragment = block.student_view(context)

        self.assertEqual(fragment.content, self.original_fragement.content)
        self.assertEqual(len(fragment.resources), 3)
        for resource in fragment.resources:
            self.assertIn(resource.mimetype, ("text/css", "application/javascript"))



"""
Utility functions for SkillTaggingMixin
"""
import requests


def get_api_client(user):
    """
    Get the authenticated api client for user
    """
    # pylint: disable=import-error,import-outside-toplevel
    from edx_rest_api_client.auth import SuppliedJwtAuth
    from openedx.core.djangoapps.oauth_dispatch.jwt import create_jwt_for_user

    client = requests.Session()
    jwt = create_jwt_for_user(user)
    client.auth = SuppliedJwtAuth(jwt)
    return client



"""
skill_tagging signal handlers
"""

from django.conf import settings
from openedx_events.event_bus import get_producer
from openedx_events.learning.signals import XBLOCK_SKILL_VERIFIED


def listen_for_xblock_skill_verified(**kwargs):
    """
    Publish openedx-event XBLOCK_SKILL_VERIFIED signal onto the event bus.
    """
    topic = getattr(settings, "EVENT_BUS_XBLOCK_VERIFICATION_TOPIC", "learning-xblock-skill-verified")
    get_producer().send(
        signal=XBLOCK_SKILL_VERIFIED,
        topic=topic,
        event_key_field='xblock_info.usage_key',
        event_data={'xblock_info': kwargs['xblock_info']},
        event_metadata=kwargs['metadata'],
    )
